# Extend Xotcl Library for XotclIDE purposes
# All Classes can belong (be defined) by an Component 
# (or can be extendet by Component)
# There is same Default Component "default" for all Classes,
# which does not belong to any Component
# Class and Object are extended with methods for handlig
# additional metainformation as Component and Method Categories

if {![xotcl::Object isobject Object]} {
   namespace import xotcl::*
}

if {$tcl_platform(platform)=="windows"} {
   set ad [pwd]
   cd $tcl_library
   package require XOTcl
   package require Tk
   cd $ad
}

namespace eval ide {
    proc lremove {list_ref elem} {
	upvar $list_ref list
	if {[set index [lsearch $list $elem]]>=0} {
	    set list [lreplace $list $index $index]
	    return 1
	}
	return 0
    }
    proc lremoveAll {list_ref listr} {
	upvar $list_ref list
        foreach elem $listr {
            lremove list $elem
        }
    }
    proc lcontain {list elem} {
	return [expr [lsearch $list $elem]>=0]
    }
    proc lappendIfNone {list_ref elem} {
	upvar $list_ref list
	if {![lcontain $list $elem]} {
	    lappend list $elem
	    return 1
	}
	return 0
    }
    proc lcollect {var_ref list test} {
	upvar $var_ref var
        set ret {}
        foreach a $list {
            set var $a
	    uplevel $test
            lappend ret $var
	}
        return $ret
    }
    proc lselect {var_ref list test} {
	upvar $var_ref var
        set ret {}
        foreach a $list {
            set var $a
            set rtest [uplevel [list expr $test]]
	    if $rtest {
                lappend ret $var
            }
        }
        return $ret
    }
    proc ldetect {var_ref list test} {
	upvar $var_ref var
        foreach a $list {
            set var $a
            set rtest [uplevel [list expr $test]]
	    if $rtest {
                return $a
            }
        }
        return
    }
}

Class instproc addCategory {category} {
    return [[self] addCategory $category {}]
}
Class instproc moveToCategory {method category} {
    return [[self] moveToCategoryB $method $category {}]
}
Class instproc renameCategory {oldname newname} {
    return [[self] renameCategoryB $oldname $newname {}]
}
Class instproc deleteCategory {name} {
    return [[self] deleteCategoryB $name {}]
}
Class instproc uncategoryFor {name} {
    return [[self] uncategoryForB $name {}]
}
Class instproc getCategoryForMethod {method} {
    return [[self] getCategoryForMethodB  $method {}]
}
Class instproc getCategories {} {
    return [[self] getCategoriesB {}]
}
Class instproc getMethodsForCategory {category} {
    return [[self] getMethodsForCategoryB $category {}]
}
Class instproc moveToComponent {app} {
    set oldapp [[self] getCompObject]
    $oldapp removeClass [self]
    [self] setMetadata component $app
    set newapp [IDE::Component getCompObjectForName $app]
    $newapp addClass [self]
}
Class instproc getHeritage {} {
    set hlist [self]
    foreach sclass [[self] info superclass] {
	if {$sclass=="::Object" || $sclass=="::xotcl::Object"} continue
	lappend hlist [$sclass getHeritage]
    }
    return [list $hlist]
}
Class instproc getChildrenHierarchy {} {
    set hlist [self]
    foreach sclass [[self] info subclass] {
	lappend hlist [$sclass getChildrenHierarchy]
    }
    return [list $hlist]
}
Class instproc getDeepChildren {} {
    set list {}
    foreach sclass [[self] info subclass] {
        set list [concat $list [$sclass getDeepChildren]]
    }
    return $list 
}
Class instproc getAllInstMethods { {ignoreList {}} } {
    set methods [[self] info instprocs]
    foreach class [[self] info heritage] {
        if {$ignoreList!="" && [ide::lcontain $ignoreList $class]} continue
	set methods [concat $methods [$class info instprocs]]
    }
    return [lsort -unique $methods]
}
Class instproc getAllFullInstMethods { {ignoreList {}} } {
    set fmethods {}
    foreach m [[self] info instprocs] {
	lappend fmethods [list $m [self]]
        set marr($m) 1
    }
    foreach class [[self] info heritage] {
        if {$ignoreList!="" && [ide::lcontain $ignoreList $class]} continue
        foreach m [$class info instprocs] {
	    if {![::info exist marr($m)]} {
                lappend fmethods [list $m $class]
                set marr($m) 1
            }
        }
    }
    return $fmethods
}

Object instproc getSubobjectsHierarchy {} {
    ::set hlist [self]
    foreach sobject [[self] info children] {
	::lappend hlist [$sobject getSubobjectsHierarchy]
    }
    return [list $hlist]
}
Object instproc inspect {} {
    [self] basicInspect
}
Object instproc basicInspect {} {
    IDE::ObjectBrowser newBrowser [self]
}
Object instproc halt {{symbol {}}} {
    if {[Object isobject IDE::Debugger]} {
       IDE::Debugger startDebugging
    }
}
Object instproc metadataDefBody {meta} {
    ::set script {}
    ::append script "@ [self] idemeta $meta [list [[self] getMetadata $meta]]\n"
    return $script
}
Object instproc metadataBody {meta} {
    return "@ [string trimleft [self] :] idemeta $meta [list [[self] getMetadata $meta]]\n"
}

Object instproc metadataAsScript {} {
    ::set script {}
    [self] instvar _idemeta
    foreach meta [array names _idemeta] {
        if {[ide::lcontain {categoriesMethodsProcs  categoriesProcs categoriesMethods categories} $meta] && $_idemeta($meta)==""} continue
	::append script [[self] metadataDefBody $meta]
    }
    return $script
}
Object instproc metadataAsScriptPur {} {
    ::set script {}
    [self] instvar _idemeta
    foreach meta [array names _idemeta] {
        if {[ide::lcontain {categoriesMethodsProcs  categoriesProcs categoriesMethods categories component} $meta]} continue
	::append script [[self] metadataDefBody $meta]
    }
    return $script
}
Object instproc getComponentName {} {
    if {[[self] hasMetadataNotEmpty component]} {
	return [[self] getMetadata component]
    }  else {
	return default
    }
}
Object instproc getCompObject {} {
    if {[[self] hasMetadataNotEmpty component]} {
	return [IDE::Component getCompObjectForName [[self] getMetadata component]]
    }  else {
	return [IDE::Component getCompObjectForName default]
    }
}
Object instproc moveToComponent {app} {
    ::set oldapp [[self] getCompObject]
    $oldapp removeObject [self]
    [self] setMetadata component $app
    ::set newapp [IDE::Component getCompObjectForName $app]
    $newapp addObject [self]
}
Object instproc renameCategoryB {oldname newname {postFix Procs}} {
    if {![[self] hasMetadata categories$postFix]} {return 0}
    ::set categories [[self] getMetadata categories$postFix]
    if {[::set index [lsearch $categories $oldname]]<0} {
	return 0
    }
    ::set categories [lreplace $categories $index $index $newname]
    [self] setMetadata categories$postFix $categories
    return 1
}
Object instproc deleteCategoryB {name {postFix Procs}} {
    if {![[self] hasMetadata categories$postFix]} {return 0}
    ::set categories [[self] getMetadata categories$postFix]
    if {[::set index [lsearch $categories $name]]<0} {
	return 0
    }
    ::set categories [lreplace $categories $index $index]
    [self] setMetadata categories$postFix $categories
    [self] setMetadata categoriesMethods$postFix [lreplace [[self] getMetadata categoriesMethods$postFix] $index $index]
    return 1
}
Object instproc addCategoryB {category {postFix Procs}} {
    [self] initCategories $postFix

    ::set categories [[self] getMetadata categories$postFix]
    if {[::set index [lsearch $categories $category]]<0} {
	::set index [llength $categories]
	::lappend categories $category
	[self] setMetadata categories$postFix $categories
	[self] setMetadata categoriesMethods$postFix [concat [[self] getMetadata categoriesMethods$postFix] [list {}]]
    }
    return $index
}
Object instproc moveToCategoryB {method category {postFix Procs}} {
    [self] uncategoryForB $method $postFix

    if {$category=="_all_categories" || $category=="_uncategorized"} return
    ::set catindex [[self] addCategoryB $category $postFix]

    ::set cm [[self] getMetadata categoriesMethods$postFix]
    ::set category [concat [lindex $cm $catindex] $method]
    ::set cm [lreplace $cm $catindex $catindex $category]
    [self] setMetadata categoriesMethods$postFix $cm
}
Object instproc uncategoryForB {method {postFix Procs}} {
    [self] initCategories $postFix

    ::set catindex 0
    foreach cat [::set cm [[self] getMetadata categoriesMethods$postFix]] {
	if {[::set index [lsearch $cat $method]]>=0} {
	    ::set cat [lreplace $cat $index $index]
	    [self] setMetadata categoriesMethods$postFix [lreplace $cm $catindex $catindex $cat]
	    return
	}
	::incr catindex
    }
}
Object instproc initCategories {postFix} {
    if {![[self] hasMetadata categories$postFix]} {
	[self] setMetadata categories$postFix {}
	[self] setMetadata categoriesMethods$postFix {}
	return
    }
}
Object instproc getCategoryForMethodB {method {postFix Procs}} {
    [self] initCategories $postFix
    
    ::set catindex 0
    foreach cat [[self] getMetadata categoriesMethods$postFix] {
	if {[::set index [lsearch $cat $method]]>=0} {
	    return [lindex [[self] getMetadata categories$postFix] $catindex]
	}
	::incr catindex
    }
    return {}
}

Object instproc getCategoriesB {{postFix Procs}} {
    [self] initCategories $postFix

    return [::concat [[self] getMetadata categories$postFix] _all_categories _uncategorized]
}
Object instproc getMethodsForCategoryB {category {postFix Procs}} {
    if {$category=="_all_categories"} {
        if {$postFix==""} {
	    ::set methods [[self] info instprocs]
	    foreach par [[self] info parameter] {
		::set parname [lindex $par 0]
		if {[::set index [::lsearch $methods $parname]]>=0} {
		    ::set methods [::lreplace $methods $index $index]
		}
	    }
	    return $methods 
	} else {
            return [[self] info procs]
        }
    } elseif {$category=="_uncategorized"} {
        ::set allcategorized {} 
        foreach c [[self] getMetadata categoriesMethods$postFix] {
            ::set allcategorized [concat $allcategorized $c]
        }
        ::set all [[self] getMethodsForCategoryB _all_categories $postFix]
        ide::lremoveAll all $allcategorized
        return $all
    }
    ::set catindex [[self] addCategoryB $category $postFix]
    return [lindex [[self] getMetadata categoriesMethods$postFix] $catindex]
}
Object instproc printString {} {
    # please overweite it to specify your Object printSring
    # the return shoul short specify the object contens for
    # displaying in object inspector
    return "[self] instance of [[self] info class] [[self] info mixin]"
}
Object instproc getDescription {} {
    if {[Object isobject [self]::description]} {
        return [self]::description
    }
    return
}
Object instproc trackOn {} {
    IDE::MethodTracker registerObjectForTracking [self]
}
Object instproc trackOff {} {
    IDE::MethodTracker unregisterObjectForTracking [self]
}
Class instproc trackOn {} {
    IDE::MethodTracker registerClassForTracking [self]
}
Class instproc trackOff {} {
    IDE::MethodTracker unregisterClassForTracking [self]
}

Object create ::IDEComps
Object create ::IDEComps::xdobry

# Metadata wrapper from old versions
Object instproc metadata args {
    puts "[self] metadata $args"
    if {[llength $args]==2} {
       [self] setMetadata [lindex $args 0] [lindex $args 1]
    } else {
	puts "calling metadata [self callingclass]>>[self callingproc] by [self callingobject]"
    }
    next
}
Object instproc hasMetadata key {
    [self] exists _idemeta($key)
}
Object instproc hasMetadataNotEmpty key {
    expr {[[self] hasMetadata $key] && [[self] set _idemeta($key)]!=""}
}
Object instproc getMetadata key {
    [self] set _idemeta($key)
}
Object instproc setMetadata {key value} {
    [self] set _idemeta($key) $value
}
Class IDEMetadataAnalyzer
IDEMetadataAnalyzer instproc unknown {args} {
    if {[llength $args]==4 && [lindex $args 1]=="idemeta" && [Object isobject [lindex $args 0]]} {
        #  $object setMetadata $type $value
	[lindex $args 0] setMetadata [lindex $args 2] [lindex $args 3]
    } else {
	[self] lappend metaList $args
    }
}
IDEMetadataAnalyzer @

#
# comments (documentation) interface
#
Object instproc hasComment {} {
  IDE::CommentsContainer objectHasComment [self]
}
Object instproc getComment {} {
  IDE::CommentsContainer objectGetComment [self]
}
Object instproc setComment  {text} {
  IDE::CommentsContainer objectSetComment [self] $text
}
Object instproc hasMethodComment {type method} {
  IDE::CommentsContainer objectMethodHasComment [self] $type $method
}
Object instproc getMethodComment {type method} {
  IDE::CommentsContainer objectMethodGetComment [self] $type $method
}
Object instproc setMethodComment {type method text} {
  IDE::CommentsContainer objectMethodSetComment [self] $type $method $text
}


